/*
** Copyright  1993, 1994 by Eric M. LaFranchi.  All Rights Reserved.
**
** This software is Copyright 1993, 1994 by Eric M. LaFranchi.  Permission
** to use, copy, and freely redistributed this software in its entirety is
** hereby granted provided that the above copyright notice and this
** permission notice are retained.  This software may not be sold for profit
** or incorporated in commercial software products without the written
** permission of the author.  This software is provided "as is", the author
** nor his employer make any representation of warranty, express or implied,
** with respect to any code or other information herein.  In addition, the
** author disclaim's any liability whatsoever for any use of such code or
** other information.
** 
**+-+
**
** Module: PACK_NNTP
**
** Abstract:
**	This module contains function to operate on news articles using
**	NNTP via TCP/IP. Current these routines have been tested with
**	Multinet, UCX and TWG.
**
** Author:
**	Eric M. LaFranchi
**
** Creation Date:
**	22-MAY-1993
**
** Modifications History:
**
**	EML002		Eric M. LaFranchi	22-DEC-1993
**	Add code to dump unused part of article in the bit bucket
**	if not used.
**
**	EML001		Eric M. LaFranchi	04-NOV-1993
**	Played with ifdefs, trying to get them correct for all the versions
**	of TCP Software.
**
**-+-
*/
#include <ctype.h>			/* Standard type definitions.	     */
#include <stdio.h>			/* Standard I/O definitions.	     */
#include <stdlib.h>			/* Standard library definitions.     */
#include <string.h>			/* Standard string definitions.      */
#include <ssdef.h>			/* openVMS system service definitions*/
#include <rms.h>			/* OpenVMS rms definitions	     */

    /* socket interface include files
     */
#pragma nostandard
#ifdef UCX
#   include <types.h>
#   include <socket.h>
#   include <in.h>
#   include <inet.h>
#   include <netdb.h>
#else
#ifdef TWG
#   include "twg$common:[netdist.include.sys]types.h"
#   include "twg$common:[netdist.include.sys]socket.h"
#   include "twg$common:[netdist.include.netinet]in.h"
#   include "twg$common:[netdist.include.arpa]inet.h"
#   include "twg$common:[netdist.include]netdb.h"
#else
#ifdef	MULTINET
#   include "multinet_root:[multinet.include.sys]types.h"
#   include "multinet_root:[multinet.include.sys]socket.h"
#   include "multinet_root:[multinet.include.netinet]in.h"
#   include "multinet_root:[multinet.include.arpa]inet.h"
#   include "multinet_root:[multinet.include]netdb.h"
    int socket( int af, int mess_type, int prot_type );
    int connect( int sd, struct sockaddr *name, int namelen );
    int send( int sd, char * msg, int length, int flags );
    int recv( int sd, char * buf, int length, int flags );
    int shutdown( int sd, int mode );
    struct hostent *gethostbyname( char *host );
    struct servent *getservbyname( char *name, char *proto );
#else
#ifdef TCPWARE
#   include "tcpware_include:types.h"
#   include "tcpware_include:socket.h"
#   include "tcpware_include:in.h"
#   include "tcpware_include:inet.h"
#   include "tcpware_include:netdb.h"
#   define send socket_send
#   define recv socket_recv
    int socket( int af, int mess_type, int prot_type );
    int connect( int sd, struct sockaddr *name, int namelen );
    int send( int sd, char * msg, int length, int flags );
    int recv( int sd, char * buf, int length, int flags );
    int shutdown( int sd, int mode );
    struct hostent *gethostbyname( char *host );
    struct servent *getservbyname( char *name, char *proto );
#else
#ifdef FUSION
#   include <flip.h>
#   include <socket.h>
#   include <in.h>
#   include <netdb.h>
#   define sockaddr_in	 in_sockaddr
#   define gethostbyname ghbyname
#   define gethostbyname gsbyname
#else /* by default use CMU, because it links */
#   include <descrip.h>
#   include <ssdef.h>
#   include <iodef.h>
#   define TCP$OPEN IO$_CREATE
#   define TCP$READ IO$_READVBLK
#   define TCP$SEND IO$_WRITEVBLK
#   define TCP$CLOSE IO$_DELETE
#   define NNTP_C_LBUFSIZ 4096
    static const $DESCRIPTOR( cmudev, "IP0:" );
#endif
#endif
#endif
#endif
#endif
#pragma standard

#include "packdef.h"			/* local const and macro definitions */
#include "vmsdef.h"			/* local VMS macro definitions	     */
#include "msgdef.h"			/* local message definitions	     */

#define NNTP_PORT	119
#define NNTP_C_BUFSIZ	512
#ifndef NNTP_C_LBUFSIZ
#define NNTP_C_LBUFSIZ	16384
#endif

struct context
{
    int s;
    size_t len;
    char *bptr;
    char buffer[NNTP_C_LBUFSIZ];
    char server[128];
    char newsgroup[128];
    size_t count, first, last;
    struct FAB fab;
    struct RAB rab;
};

#ifdef VAXC
#pragma nostandard
#endif

extern size_t (*pack_connect)( void **const, const char *const );
extern size_t (*select_articles)( void *const, const char *const );
extern size_t (*scan_articles)( void *const, size_t (*)(), const void * );
extern size_t (*extract_articles)( void *const, const size_t, size_t (*)(), const void * );
extern size_t (*cleanup_articles)( void *const, const size_t );
extern size_t (*disconnect)( void **const );

extern const unsigned int gblflags;

#ifdef VAXC
#pragma standard
#endif

static size_t (*nntp_read_line)( void * cxt, char *const buffer );
static size_t (*nntp_send_line)( void * cxt, const char *const buffer );

    /* If you need to specify username and password for DECNET logins.
     * Specifies as follows:
     *
     *  #define AUTHORIZATION "\"user password\""
     *
     * The username and password can be specified on the command line
     * as follows:
     *
     *	cc/define=(authorization="""\""user password\""""") pack_nntp.c ...
     */
#ifndef AUTHORIZATION 
#define AUTHORIZATION ""
#endif

#ifdef htons
#undef htons
#endif

#define htons( p )	((((p) << 8) & 0xff00) | (((p) >> 8) & 0x00ff))

static const char tcp[ ] = "tcp";
static const char nntp[ ] = "nntp";
static const char authorization[ ] = AUTHORIZATION;


/*
 *+-+
 *
 * Function: nntp_tcp_read_line
 *
 * Abstract:
 *	returns one line of input from the nntp server. This routine
 *	maintains a pointer and length to the last buffer read from the
 *	nntp server. If the buffer length is zero or only part of a line
 *	exists, then another read is posted to the nntp server;
 *
 * Inputs:
 *	cxt -- pointer NNTP context block
 *
 *	input -- pointer to user buffer to recieve next input line.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_tcp_read_line( void *context,
		    char *const buf )
{
    IOSB iosb;

    register status;
    register char *p;
    register size_t len, idx;
    register struct context *cxt;

    cxt = (struct context *)context;

    if ( (cxt->len == 0) || ((p = strchr( cxt->bptr, '\012' )) == NULL) )
    {
	if ( cxt->len > 0 )
	    strcpy( cxt->buffer, cxt->bptr );

	do
	{
	    idx = cxt->len;
	    len = (sizeof( cxt->buffer ) - 1) - idx;

#		if defined ( MULTINET ) || defined ( TCPWARE ) || defined ( FUSION ) || defined ( TWG ) || defined ( UCX ) 

		/* Read in the next buffer from the NNTP server
		 */
	    len = recv( cxt->s, &cxt->buffer[idx], len, 0 );
	    if ( len == -1 )
		raise_exception( PACKASM_RECVERR, 1, cxt->server );

#		else /* CMU */

		/* Read in the next buffer from the NNTP server
		 */
	    status = SYS$QIOW( 0, cxt->s, TCP$READ, &iosb, 0, 0,
			       &cxt->buffer[idx], len, 0, 0, 0, 0);
	    if ( status & 1 ) status = iosb.status;
	    if ( !(status & 1) )
		raise_exception( PACKASM_RECVERR, 1, cxt->server, status );
	    len = iosb.bytcnt;
#		endif /* CMU */

	    cxt->len += len;
	    cxt->bptr = cxt->buffer;
	    cxt->buffer[cxt->len] = '\0';
	} while ( (p = strchr( cxt->bptr, '\012' )) == NULL );
    }

	/* write Null character over CR and advance past the LF
	 */
    *--p = '\0';
    p += 2;
    strcpy( buf, cxt->bptr );
    cxt->len -= (p - cxt->bptr);
    cxt->bptr = p;

    return ( PACKASM_SUCCESS );
}

/*
 *+-+
 *
 * Function: nntp_dnet_read_line
 *
 * Abstract:
 *	returns one line of input from the nntp server. This routine
 *	maintains a pointer and length to the last buffer read from the
 *	nntp server. If the buffer length is zero or only part of a line
 *	exists, then another read is posted to the nntp server;
 *
 * Inputs:
 *	cxt -- pointer NNTP context block
 *
 *	input -- pointer to user buffer to recieve next input line.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_dnet_read_line( void *context,
		     char *const buf )
{
    IOSB iosb;

    register status;
    register char *p;
    register size_t len, idx;
    register struct context *cxt;

    cxt = (struct context *)context;

    if ( !((cxt->len != 0) && ((p = strchr( cxt->bptr, '\012' )) != NULL)) )
    {
	if ( cxt->len > 0 )
	    strcpy( cxt->buffer, cxt->bptr );

	do
	{
	    idx = cxt->len;
	    len = (sizeof( cxt->buffer ) - 1) - idx;

	    cxt->rab.rab$l_ubf = &cxt->buffer[idx];
	    cxt->rab.rab$w_usz = len;
	    status = SYS$GET( &cxt->rab );
	    if ( !(status & 1) )
		raise_exception( PACKASM_RECVERR, 1, cxt->server, status );
	    len = cxt->rab.rab$w_rsz;;

	    cxt->len += len;
	    cxt->bptr = cxt->buffer;
	    cxt->buffer[cxt->len] = '\0';
	} while ( (p = strchr( cxt->bptr, '\012' )) == NULL );
    }

	/* write Null character over CR and advance past the LF
	 */
    *--p = '\0';
    p += 2;
    strcpy( buf, cxt->bptr );
    cxt->len -= (p - cxt->bptr);
    cxt->bptr = p;

    return ( PACKASM_SUCCESS );
}

/*
 *+-+
 *
 * Function: nntp_tcp_send_line
 *
 * Abstract:
 *	sends one line of input to the nntp server
 *
 * Inputs:
 *	cxt -- pointer NNTP context block
 *
 *	sndbuf	-- pointer to the buffer to be sent to the nntp server
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_tcp_send_line( void *context,
		    const char *const msgbuf )
{
    IOSB iosb;
    char sndbuf[NNTP_C_BUFSIZ];

    register status;
    register struct context *cxt;

    cxt = (struct context *)context;

	/* add carriage return to the end of the line being sent
	 */
    sprintf( sndbuf, "%s\015\012", msgbuf );

#	    if defined ( MULTINET ) || defined ( TCPWARE ) || defined ( FUSION ) || defined ( TWG ) || defined ( UCX ) 

	/* send specified line to the NNTP server
	 */
    if ( send( cxt->s, sndbuf, strlen( sndbuf ), 0 ) < 0 )
	raise_exception( PACKASM_SENDERR, 1, cxt->server );
#	    else /* CMU */

	/* send specified line to the NNTP server
	 */
    status = SYS$QIOW ( 0, cxt->s, TCP$SEND, &iosb, NULL, NULL,
			sndbuf, strlen( sndbuf ), 0, 0, 0, 0);
    if ( status & 1 ) status = iosb.status;
    if ( !(status & 1) )
	raise_exception( PACKASM_SENDERR, 1, cxt->server, status );

#	    endif /* CMU */

    return ( PACKASM_SUCCESS );
}

/*
 *+-+
 *
 * Function: nntp_dnet_send_line
 *
 * Abstract:
 *	returns one line of input from the nntp server. This routine
 *	maintains a pointer and length to the last buffer read from the
 *	nntp server. If the buffer length is zero or only part of a line
 *	exists, then another read is posted to the nntp server;
 *
 * Inputs:
 *	cxt -- pointer NNTP context block
 *
 *	input -- pointer to user buffer to recieve next input line.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_dnet_send_line( void *context,
		     const char *const msgbuf )
{
    IOSB iosb;
    char sndbuf[NNTP_C_BUFSIZ];

    register status;
    register struct context *cxt;

    cxt = (struct context *)context;

	/* add carriage return to the end of the line being sent
	 */
    sprintf( sndbuf, "%s\015\012", msgbuf );

    cxt->rab.rab$l_rbf = sndbuf;
    cxt->rab.rab$w_rsz = strlen( sndbuf );
    status = SYS$PUT( &cxt->rab );
    if ( !(status & 1) )
	raise_exception( PACKASM_SENDERR, 1, cxt->server, status );

    status = SYS$FLUSH( &cxt->rab );
    if ( !(status & 1) )
	raise_exception( PACKASM_SENDERR, 1, cxt->server, status );

    return ( PACKASM_SUCCESS );
}


/*
 *+-+
 *
 * Function: nntp_tcp_connect
 *
 * Abstract:
 *
 * Inputs:
 *	context -- pointer to a void pointer in which a context block can
 *		   and accessed by these routines.
 *
 *	server -- pointer to a character string containing the server name.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_tcp_connect( void **const context,
	      const char *const server )
{
    char	buf[NNTP_C_BUFSIZ], *p;
    char	hostaddr[NNTP_C_BUFSIZ];
#	if defined ( MULTINET ) || defined ( TCPWARE ) || defined ( FUSION ) || defined ( TWG ) || defined ( UCX ) 
    struct	hostent *host;
    struct	servent *serv;
    struct	sockaddr_in sin;
#	else
    IOSB	iosb;
#	endif

    register struct context *cxt;
    register status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)calloc( 1, sizeof( *cxt ) );
    if ( cxt == NULL )
	raise_exception( PACKASM_INSVIRMEM, 0 );

#	if defined ( MULTINET ) || defined ( TCPWARE ) || defined ( FUSION ) || defined ( TWG ) || defined ( UCX ) 

	/* Get the port number for the "nntp" service
	 * Because UCX is brain dead, check -1 as a failure for
	 * getservbyname. WOW!
	 */
    memset( (char *)&sin, 0, sizeof( sin ) );
    serv = getservbyname( (char *)nntp, (char *)tcp );
    if ( (serv == NULL) || (serv = (struct servent *)(-1)) )
#	    ifdef NNTP_PORT
	sin.sin_port = htons( NNTP_PORT );
#	    else
	raise_exception( PACKASM_CONNECTERR, 1, server,
			 PACKASM_UNKNOWNSERV, 1, nntp );
#	    endif
    else
	sin.sin_port = serv->s_port;

	/* search host database for host name.
	 */
    host = gethostbyname( (char *)server );
    if ( host == NULL )
    {
	    /* check for host address of the form [x.x.x.x] and strip
	     * off the braces, so we only pass addresses of the form
	     * x.x.x.x
	     */
	if ( server[0] == '[' )
	{
	    strcpy( hostaddr, &server[1] );
	    hostaddr[strlen( hostaddr ) - 1] = '\0';
	}
	else
	    strcpy( hostaddr, server );

 	    /* host name not found, assume internet style address.
	     */
#ifdef TCPWARE
	sin.sin_addr = inet_addr( (char *)hostaddr );
#else
	sin.sin_addr.s_addr = inet_addr( (char *)hostaddr );
#endif
	if ( sin.sin_addr.s_addr == -1)
	    raise_exception( PACKASM_CONNECTERR, 1, server,
			     PACKASM_UNKNOWNHOST, 1, server );
	sin.sin_family = AF_INET;
    }
    else
    {
	sin.sin_family = host->h_addrtype;
	memcpy( &sin.sin_addr, host->h_addr, host->h_length );
    }

	/* create a communication endpoint for TCP/IP socket
	 * services
	 */
    cxt->s = socket( AF_INET, SOCK_STREAM, 0 );
    if ( cxt->s < 0 )
	raise_exception( PACKASM_CONNECTERR, 1, server,
			 PACKASM_SOCKETERR, 0, SS$_NOSUCHDEV );

	/* create connection to the specified server.
	 */
    status = connect( cxt->s, (struct sockaddr *)&sin, sizeof(sin) );
    if ( status < 0 )
	raise_exception( PACKASM_CONNECTERR, 1, server,
			 PACKASM_CONNECTFAIL, 0 );

#	else /* CMU */

	/* Assign a channel to the CMU device
	 */
    cxt->s = 0;
    status = SYS$ASSIGN( &cmudev, (unsigned short *)&cxt->s, NULL, NULL );
    if ( !(status & 1) )
	raise_exception( PACKASM_CONNECTERR, 1, server,
			 PACKASM_CONNECTFAIL, 0, status );

	/* create connection to the specified server.
	 */
    status = SYS$QIOW( 0, cxt->s, TCP$OPEN, &iosb, NULL, NULL,
		       server, NNTP_PORT, 0, 1, 0, 0 );
    if ( status & 1 ) status = iosb.status;
    if ( !(status & 1) )
	raise_exception( PACKASM_CONNECTERR, 1, server,
			 PACKASM_CONNECTFAIL, 0, status );
#	endif /* CMU */

	/* get response from nntp server
	 */
    status = nntp_read_line( cxt, buf );
    if ( !(status & 1) )
	return ( status );

	/* check status and display server name to user.
	 */
    status = atol( buf );
    if ( (status != 200) && (status != 201) )
	raise_exception( PACKASM_CONNECTERR, 1, server,
			 PACKASM_INVSERVRESP, 0 );

    display( PACKASM_CONNECT, 1, &buf[4] );

	/* save the server name in out context block
	 */
    strcpy( cxt->server, server );

	/* load context block pointer into external context block.
	 */
    *context = cxt;

    return ( PACKASM_SUCCESS );
}

/*
 *+-+
 *
 * Function: nntp_dnet_connect
 *
 * Abstract:
 *
 * Inputs:
 *	context -- pointer to a void pointer in which a context block can
 *		   and accessed by these routines.
 *
 *	server -- pointer to a character string containing the server name.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_dnet_connect( void **const context,
		   const char *const server )
{
    static const char NNTP_object[] = "::\"0=NNTP\"";
    char buf[NNTP_C_BUFSIZ];

    register status;
    register char *p;
    register struct context *cxt;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)calloc( 1, sizeof( *cxt ) );
    if ( cxt == NULL )
	raise_exception( PACKASM_INSVIRMEM, 0 );

	/* account for colons that may exist on a DECNET node name
	 */
    if ( p = strstr( server, "::" ) )
	*p == '\0';

	/* create decnet task name
	 */
    sprintf( buf, "%s%s%s", server, authorization, NNTP_object );

	/* set up the RMS data structures
	 */
    cxt->fab = cc$rms_fab;
    cxt->fab.fab$v_put = 1;
    cxt->fab.fab$v_get = 1;
    cxt->fab.fab$l_fna = buf;
    cxt->fab.fab$b_fns = strlen( buf );

	/* open DECNET connection to NNTP server
	 */
    status = SYS$OPEN( &cxt->fab );
    if ( !(status & 1) )
	raise_exception( PACKASM_CONNECTERR, 1, server,
			 cxt->fab.fab$l_sts, cxt->fab.fab$l_stv );

	/* Establish a stream to the NNTP server
	 */
    cxt->rab = cc$rms_rab;
    cxt->rab.rab$l_fab = &cxt->fab;
    status = SYS$CONNECT( &cxt->rab );
    if ( !(status & 1) )
    {
	SYS$CLOSE( &cxt->fab );
	raise_exception( PACKASM_CONNECTERR, 1, server, status );
    }

	/* get response from nntp server
	 */
    status = nntp_read_line( cxt, buf );
    if ( !(status & 1) )
	return ( status );

	/* check status and display server name to user.
	 */
    status = atol( buf );
    if ( (status != 200) && (status != 201) )
	raise_exception( PACKASM_CONNECTERR, 1, server );

    display( PACKASM_CONNECT, 1, &buf[4] );

	/* save the server name in our context block
	 */
    strcpy( cxt->server, server );

	/* load context block pointer into external context block.
	 */
    *context = cxt;

    return ( PACKASM_SUCCESS );
}


/*
 *+-+
 *
 * Function: nntp_select_newsgroup
 *
 * Abstract:
 *	This function selects the specified newsgroup
 *
 * Inputs:
 *	context -- pointer to a void pointer which contains the NNTP context
 *		   block.
 *
 *	newsgroup -- pointer to newsgroup.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_select_newsgroup( void *const context,
		       const char *const newsgroup )
{
    static char group_select[NNTP_C_BUFSIZ];

    char buf[NNTP_C_BUFSIZ], *q;
    char const *p;

    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)context;

	/* convert to lower case, while formatting request
	 */
    strcpy( group_select, "GROUP " );
    q = &group_select[strlen( group_select )];
    for ( p = newsgroup; *p; p++ )
	*q++ = _tolower( *p );
    *q = '\0';

	/* send line to nntp server
	 */
    status = nntp_send_line( cxt, group_select );
    if ( !(status & 1) )
	return ( status );

	/* get response from nntp server
	 */
    status = nntp_read_line( context, buf );
    if ( !(status & 1) )
	return ( status );

	/* Interpret status code from nntp server
	 */
    status = atol( buf );
    if ( status == 411 )
	raise_exception( PACKASM_NOTEXIST, 1, newsgroup );

    sscanf( &buf[4], "%d %d %d %s", &cxt->count, &cxt->first, &cxt->last, buf );

    strcpy( cxt->newsgroup, newsgroup );

    return ( PACKASM_SUCCESS );
}


/*
 *+-+
 *
 * Function: nntp_scan_articles
 *
 * Abstract:
 *	This routine scans the specified newsgroups for subject lines
 *	that have names that match the specified package name. If a match
 *	is found the newsgroup and article number are passed to the
 *	add_part_record function.
 *
 * Inputs:
 *	context -- pointer to a void pointer which contains the NNTP context
 *		   block.
 *	callback -- pointer to function to process subject
 *	callarg -- pointer to context block passed to callback function
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 * Special Notes
 *	This routine uses XHDR command which is not part of the rfc977
 *	standard, but is implemented by the Berkeley NNTP server.
 *-+-
 */
static size_t
nntp_scan_articles( void *const context,
		    size_t (*const callback)( ),
		    const void *callarg )
{
    char buf[NNTP_C_BUFSIZ];
    char header[NNTP_C_BUFSIZ];
    char subject[NNTP_C_BUFSIZ];
    size_t code, artnum;
    size_t msgid, sublen, *date, datebuf[2];

    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)context;

#	ifndef	NOXHDR
    (void) sprintf( header, "XHDR subject %ld-%ld", cxt->first, cxt->last );

	/* request article header from the NNTP server
	 */
    status = nntp_send_line( cxt, header );
    if ( !(status & 1) )
	return ( status );

	/* get response from nntp server
	 */
    status = nntp_read_line( context, buf );
    if ( !(status & 1) )
	return ( status );

    code = atol( buf );
    if ( code >= 300 )
	raise_exception( PACKASM_NOXHDRSUP, 1, cxt->server );

    date = NULL;

	/* read header looking for subject and date lines
	 */
    for ( ;; )
    {
	status = nntp_read_line( context, buf );
	if ( !(status & 1) )
	    return ( status );

	    /* check for last subject
	     */
	if ( (strlen( buf ) == 1) && (buf[0] == '.') )
	    break;

	    /* parse returned subject line
	     */
	msgid = atol( buf );
	strcpy( subject, (strchr( buf, ' ' ) + 1) );


#	else /* NOXHDR */

    date = datebuf;

	/* select each article in the newsgroup, then try to determine
	 * if it this part of a package we are trying to extract.
	 */
    for ( msgid = cxt->first; msgid <= cxt->last; msgid++ )
    {
	sprintf( header, "HEAD %d", msgid );

	    /* request article body from the NNTP server
	     */
	status = nntp_send_line( cxt, header );
	if ( !(status & 1) )
	    return ( status );

	    /* get response from nntp server
	     */
	status = nntp_read_line( context, buf );
	if ( !(status & 1) )
	    return ( status );

	sscanf( buf, "%d %d", &code, &artnum );

	    /* if article doesn't exist continue
	     */
	if ( code == 423 )
	    continue;

	if ( (code != 221) || (artnum != msgid) )
	    raise_exception( PACKASM_NOSUCHART, 3, msgid,
			     code, cxt->newsgroup );

	    /* read header looking for subject and date lines
	     */
	for ( ;; )
	{
	    status = nntp_read_line( context, buf );
	    if ( !(status & 1) )
		return ( status );

	    if ( (strlen( buf ) == 1) && (buf[0] == '.') )
		break;

	    if ( strncmp( buf, "Subject: ", 9 ) == 0 )
	    {
		strcpy( subject, &buf[9] );
		continue;
	    }
	}

#	endif	/* NOXHDR */

	    /* process subject line
	     */
	status = callback( callarg, subject, msgid, date );
	if ( !(status & 1) )
	    continue;
    }
    return ( PACKASM_SUCCESS );
}


/*
 *+-+
 *
 * Function: nntp_extract_article
 *
 * Abstract:
 *	This function locates specified part using the newsgroup and msgid
 *	information specified by the user. and then calls a callback record
 *	for each text record.
 *
 * Inputs:
 *
 * Outputs:
 *	None
 *
 * Returns:                                   
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_extract_article( void *const context,
		      const unsigned int msgid,
		      size_t (*const callback)( ),
		      const void *callarg )
{
    size_t code, artnum;
    char article_body[NNTP_C_BUFSIZ];
    char buf[NNTP_C_BUFSIZ];

    register struct context *cxt;
    register size_t status, retstat;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)context;

    sprintf( article_body, "BODY %d", msgid );

	/* request article body from the NNTP server
	 */
    status = nntp_send_line( cxt, article_body );
    if ( !(status & 1) )
	return ( status );

	/* get response from nntp server
	 */
    status = nntp_read_line( context, buf );
    if ( !(status & 1) )
	return ( status );

    sscanf( buf, "%d %d", &code, &artnum );

    if ( (code != 222) || (artnum != msgid) )
	raise_exception( PACKASM_NOSUCHART, 3, msgid,
			 code, cxt->newsgroup );

    retstat = PACKASM_SUCCESS;			/* EML002 */

	/* read text records from the NNTP server and pass to the
	 * specified callback routine.
	 */
    for ( ;; )
    {
	status = nntp_read_line( context, buf );
	if ( !(status & 1) )
	    return ( status );

	    /* must be done.
	     */
	if ( (strlen( buf ) == 1) && (buf[0] == '.') )
	    break;

	if ( retstat == PACKASM_SUCCESS )	/* EML002 */
	{
	    status = callback( callarg, buf );
	    if ( !(status & 1) )
		retstat = status;
	}
    }

    return ( retstat );
}

/*
 *+-+
 *
 * Function: nntp_extract_article
 *
 * Abstract:
 *	This function locates specified part using the newsgroup and msgid
 *	information specified by the user. and then calls a callback record
 *	for each text record.
 *
 * Inputs:
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_cleanup( void *const context,
	      const unsigned int msgid )
{
    unsigned int record_length;
    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)context;

    return ( PACKASM_SUCCESS );
}

/*
 *+-+
 *
 * Function: nntp_tcp_disconnect
 *
 * Abstract:
 *	nntp_disconnect contains code to shut down an open TCP/IP connection
 *	with an NNTP_SERVER.
 *
 * Inputs:
 *	context -- pointer to a void pointer which contains the NNTP context
 *		   block.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_tcp_disconnect( void **const context )
{
    char buf[NNTP_C_BUFSIZ];

    register struct context *cxt;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)*context;

	/* notify the server to shutdown the connect on
	 * both ends.
	 */
    nntp_send_line( cxt, "QUIT" );

#	if defined ( MULTINET ) || defined ( TCPWARE ) || defined ( FUSION ) || defined ( TWG ) || defined ( UCX ) 
    shutdown( cxt->s, 2 );
#	else
    (void)SYS$QIOW( 0, cxt->s, TCP$CLOSE, NULL, NULL, NULL,
	      2, 0, 0, 0, 0, 0 );
    (void)SYS$DASSGN( cxt->s );
#	endif

	/* free context block and set pointer to NULL
	 */
    *context = NULL;
    free( cxt );

    return ( PACKASM_SUCCESS );
}

/*
 *+-+
 *
 * Function: nntp_dnet_disconnect
 *
 * Abstract:
 *	nntp_dnet_disconnect terminates the session with the NNTP sever
 *
 * Inputs:
 *	context -- pointer to a void pointer which contains the NNTP context
 *		   block.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
static size_t
nntp_dnet_disconnect( void **const context )
{
    char buf[NNTP_C_BUFSIZ];

    register struct context *cxt;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)*context;

	/* notify the server to shutdown the connect on
	 * both ends.
	 */
    nntp_send_line( cxt, "QUIT" );

    SYS$CLOSE( &cxt->fab );

	/* free context block and set pointer to NULL
	 */
    *context = NULL;
    free( cxt );

    return ( PACKASM_SUCCESS );
}

/*
 *+-+
 *
 * Function: load_nntp_tcp_addresses
 *
 * Abstract:
 *	this routine loads the addresses of the tcp based NNTP routines
 *	into global function varaibles.
 *
 * Inputs:
 *	None
 *
 * Implicit Inputs:
 *	pack_connect
 *	select_articles
 *	scan_articles
 *	extract_articles
 *	cleanup_articles
 *	disconnect
 *
 * Outputs:
 *	global function	pointer to mail routines loaded
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
size_t
load_nntp_tcp_routines( )
{
    pack_connect = nntp_tcp_connect;
    select_articles = nntp_select_newsgroup;
    scan_articles = nntp_scan_articles;
    extract_articles = nntp_extract_article;
    cleanup_articles = nntp_cleanup;
    disconnect = nntp_tcp_disconnect;
    nntp_send_line = nntp_tcp_send_line;
    nntp_read_line = nntp_tcp_read_line;

    return ( PACKASM_SUCCESS );
}

/*
 *+-+
 *
 * Function: load_nntp_decnet_addresses
 *
 * Abstract:
 *	this routine loads the addresses of the tcp based NNTP routines
 *	into global function varaibles.
 *
 * Inputs:
 *	None
 *
 * Implicit Inputs:
 *	pack_connect
 *	select_articles
 *	scan_articles
 *	extract_articles
 *	cleanup_articles
 *	disconnect
 *
 * Outputs:
 *	global function	pointer to mail routines loaded
 *
 * Returns:
 *	returns VMS status error or PACKASM_SUCCESS
 *
 *-+-
 */
size_t
load_nntp_dnet_routines( )
{
    pack_connect = nntp_dnet_connect;
    select_articles = nntp_select_newsgroup;
    scan_articles = nntp_scan_articles;
    extract_articles = nntp_extract_article;
    cleanup_articles = nntp_cleanup;
    disconnect = nntp_dnet_disconnect;
    nntp_send_line = nntp_dnet_send_line;
    nntp_read_line = nntp_dnet_read_line;

    return ( PACKASM_SUCCESS );
}
